/*
+  This is the naive method for solving the all-nearest-neighbors problem.
+  For each point, go through all others and find the one that is closest.
+
+  Copyright Michael Neidhardt, June 2008.
+
+  New version that combines the fp and rational versions.
+  T=type of coordinates, N=type of the point, i.e. the coordinate-container.
*/

#ifndef ANNBRUTEFORCE_H
#define ANNBRUTEFORCE_H

#include <iostream>
#include <cmath>
#include <vector>
#include <algorithm>
#include <utility>

using namespace std;


template<typename T, typename N>
class ANNBruteforce {
public:
   ANNBruteforce(bool dup=true);
   vector<vector<N> > findAllNearestNeighbors(vector<N>& ps);
   vector<N> findNeighbors(N& p, vector<N>& ps);

private:
   bool duplicatesAllowed;
   vector<N> pointset;
   typedef pair<T,int> mypair;
};

template<typename T, typename N>
T squared_distance(N& p1, N& p2);


/* -------- Implementation here: ---------- */

template<typename T, typename N>
ANNBruteforce<T,N>::ANNBruteforce(bool dup) : duplicatesAllowed(dup) {}



/* Finds all nearest neighbors for all points in pointset.
+  Returns a vector of Point-vectors. Each Point-vector contains
+  at least 2 Points: a point p and its nearest neighbor.
+  If p has more than 1 nearest neighbor (i.e. the same distance away),
+  those other neigboring points are appended to the Point-vector.
*/
template<typename T, typename N>
vector<vector<N> > ANNBruteforce<T,N>::findAllNearestNeighbors(vector<N>& pointset) {

   cout.setf(ios::fixed);

   vector<vector<N> > result;

   for (unsigned int i=0; i<pointset.size(); i++) {
      result.push_back(findNeighbors(pointset[i], pointset));
   }
   
   return result;
}

/* Finds all nearest neighbors for one point.
+  Returns a vector of Point's, i.e. all those other
+  points that are p's nearest neighbors.
+  This means of course that they are all the same distance away.
+  I return the complete set of nearest neighbors for a point so that
+  i can use this to check other algorithms. Other algorithms might
+  just find one NN, and to be correct, such a point must be in the
+  list returned by this method.
*/
template<typename T, typename N>
vector<N> ANNBruteforce<T,N>::findNeighbors(N& p, vector<N>& pointset) {

   vector<N> result(0);
   vector<mypair> tmpresults(0);
   // Used for temporary storage, so I can sort distances. This is needed
   // in case a point has more than one nearest neighbor. If so,
   // I want to return all those neighbor points that are the 
   // same (minimum) distance away. I then calculate distance from the
   // point p to all others, and store results in tmpresults.
   // This is sorted, and I pick out the first and all others that have 
   // the same distance.

   int samepoint_counter=0;
   unsigned int duplicateIdx=-1;
   
   for (unsigned int i=0; i<pointset.size(); i++) {
      if (p.x() != pointset[i].x() || p.y() != pointset[i].y()) {

         // T tmp = squared_distance(p, pointset[i]); 
         // Temporary solution:
         T tmp = (p.x()-pointset[i].x())*(p.x()-pointset[i].x()) + 
                 (p.y()-pointset[i].y())*(p.y()-pointset[i].y());

         tmpresults.push_back(mypair(tmp,i));
      }
      else {
         ++samepoint_counter;
         duplicateIdx = i;
      }
      /* If pointset contains duplicates, I have decided that, as default
      +  behaviour, each duplicate has itself as nearest neighbor.
      +  This can be disallowed by calling the constructor with an arg of false.
      +  On the other hand, if there are no duplicates, a point
      +  should obviously not see itself as its nearest neighbor.
      +  This variable enables me to check if there are duplicates.
      */
   }

   /* If there are duplicates of the point I'm finding NN for, 
   +  samepoint_counter is at least 2, and if so, I add 0 as distance
   +  and the id for one of the duplicates.
   */
   if (samepoint_counter > 1 && duplicatesAllowed) {
      tmpresults.push_back(mypair(0, duplicateIdx));
   }

   // Sort asc. on distance.
   sort(tmpresults.begin(), tmpresults.end());

   // The first one (i.e. at index 0) is always the nearest,
   // so add that to final result, and also any other point
   // that is the same distance away:
   unsigned int idx=0;
   do {
      result.push_back(pointset[tmpresults[idx].second]);
      ++idx;
   } while (idx < tmpresults.size() &&
            tmpresults[idx-1].first == tmpresults[idx].first);

   return result;
}

#endif
