/* A class for easily comparing two sets of points.
+  It is a means of testing one program's output against
+  that of another.
+  I've created different programs that all calculate 
+  all-nearest-neighbors for a point set. To test that my 
+  programs are correct, I made a brute force version,
+  assumed to be correct, as it is quite simple ;-)
+  This class is supposed to take the result from one
+  program, and compare each point and its NN to the result 
+  from the Bruteforce version.
+
+  This is a new version that is meant to be used for both floating point
+  and rational (exact) types. 13. august 2008.
+
+  Typenames: T = type of the coordinates.
+             N = type of the point, i.e. the coordinate-container.
+
+  Copyright Michael Neidhardt, June 2008.
*/


#include <iostream>
#include <vector>
#include "ANNBruteforce2.h"

using namespace std;

template<typename T, typename N>
class ANNComparator {
public:
   ANNComparator(bool dup=true);
   int doComparison(vector<pair<N, N> >& ann, vector<N>& pointset, int verbose);
   int doComparison(vector<pair<N, N> >& ann, int verbose);
private:
   bool isMember(N& p, vector<N >& results, int verbose);
   bool duplicatesAllowed;    // See ANNBruteforce2.h for explanation.
};

template<typename T, typename N>
ANNComparator<T,N>::ANNComparator(bool dup) : duplicatesAllowed(dup) {}


/* This checks whether results from another algorithm are identical to brute force results.
+  This check is complicated by the fact that a point might have more than one
+  nearest neighbor, i.e. the same distance away. My Brute Force algorithm finds
+  them all, where as the others only find one.
+  The arg ann contains pairs of nodes, pair(node, nn). I.e. each pair contains
+  a node and its n earest neighbor.
+  The arg pointset contains the original pointset, so that brute force can calculate
+  the nearest neighbor for each of them easily. */
template<typename T, typename N>
int ANNComparator<T,N>::doComparison(vector<pair<N, N> >& ann,
                  vector<N >& pointset, int verbose) {

   int errors=0;
   ANNBruteforce<T,N> bruteforce(duplicatesAllowed);

   for (unsigned int i=0; i<ann.size(); i++) {
      if (verbose > 1) {
         std::cout << ann[i].first.x() << "," << ann[i].first.y() << " NN="
                   << ann[i].second.x() << "," << ann[i].second.y() << "     BF: ";
      }
      
      // Find neighbors for this point via brute force:
      vector<N > bfresult = bruteforce.findNeighbors(ann[i].first, pointset);

      // Check if the neighbor found in other algorithm was also found in Bruteforce:
      if (isMember(ann[i].second, bfresult, verbose)) { 
         if (verbose > 1) cout << " OK.\n";
      }
      else {
         if (verbose > 1) cout << " ERROR.\n";
         ++errors;
      }

      //if (i>0 && i%1000==0) std::cout << "Checking " << i << ". " << errors << " errors found.\n";
   }
   
   return errors;
}

/* Basically the same as the above function, except that this one creates pointset
+  from ann, and then calls doComparison above.
+  This is because the Delaunay functions (TTL and CGAL) do things a bit
+  differently from my plane sweep algorithm. */
template<typename T, typename N>
int ANNComparator<T,N>::doComparison(vector<pair<N, N> >& ann, int verbose) {

   vector<N> pointset(0);

   for (unsigned int i=0; i<ann.size(); i++) {
      pointset.push_back(ann[i].first);
   }
   
   return doComparison(ann, pointset, verbose);
}


/* This checks if the point p is found in the vector results.
+  Returns true if so.
*/
template<typename T, typename N>
bool ANNComparator<T,N>::isMember(N& p, vector<N >& results, int verbose) {
   bool member=false;
   
   for (unsigned int j=0; j<results.size(); j++) {
      //if (p == results[j]) {   // Original version
      if (p.x() == results[j].x() && p.y() == results[j].y()) {
         member = true;
      }

      if (verbose > 1) {
         std::cout << "[" << results[j].x() << "," << results[j].y() << "] ";
      }
   }
   
   return member;
}

