/* Class that holds a point in 2D, i.e.
+  it has 2 coordinates, x and y.
+
+  Originally made for the ANN Plane sweep algorithm.
+  Copyright Michael Neidhardt, May 2008.
*/

#ifndef POINT_H
#define POINT_H

#include <iostream>
#include <string>
#include <cmath>
//#include <LEDA/numbers/rational.h>

using namespace std;

template <typename T>
class Point {
public:
   Point();
   Point(T xcoord, T ycoord);
   Point(const Point<T>& p);           // Copy constructor.

   void init(T xcoord, T ycoord) { x_ = xcoord; y_ = ycoord; }
      
   T x() const { return x_; }
   T y() const { return y_; }

   void print() const { print(""); }
   void print(std::string info) const;
   
   Point<T>& operator=(const Point<T>& rightarg);

   T& operator[](unsigned int i) { 
      if (i==0)      return x_;
      else if (i==1) return y_;
      else {
         std::cerr << "Index out of range in Point.\n"; 
         exit(1);
      }
   }
   //const T& operator[](unsigned int i) const { }

private:
   T x_;
   T y_;
};

template<typename T>
T squared_distance(const Point<T>& p, const Point<T>& q);
template<typename T>
double squared_distance_dbl(const Point<T>& p, const Point<T>& q);

/* ----- Non member stuff: -------- */
template< typename T >
bool operator == (const Point<T>& p, const Point<T>& q);
template< typename T >
bool operator != (const Point<T>& p, const Point<T>& q);
template< typename T >
bool operator<(const Point<T>& p, const Point<T>& q);
template< typename T >
bool operator>(const Point<T>& p, const Point<T>& q);


// -------- Implementation here: ----------
template< typename T >
Point<T>::Point() : x_(0), y_(0) {}

template< typename T >
Point<T>::Point(T xcoord, T ycoord) : x_(xcoord), y_(ycoord) {}

template< typename T >             // Copy constructor.
Point<T>::Point(const Point<T>& p) : x_(p.x()), y_(p.y()) {}


template< typename T >
void Point<T>::print(std::string info) const {
   std::cout << info << "[" << x_ << "," << y_ << "]";
}

template< typename T >
Point<T>& Point<T>::operator=(const Point<T>& rightarg) {
   x_ = rightarg.x();
   y_ = rightarg.y();

   return *this;
}




/* ---------- Non members here: --------- */
template< typename T >
bool operator==(const Point<T>& p, const Point<T>& q) { 
   return (p.x() == q.x() && p.y() == q.y());
}


template< typename T >
bool operator!=(const Point<T>& p, const Point<T>& q) { 
   return (p.x() != q.x() || p.y() != q.y());
}


template< typename T >
bool operator<(const Point<T>& p1, const Point<T>& p2) { 
   if (p1.x() != p2.x()) { return p1.x() < p2.x(); }
   else return p1.y() < p2.y();
}

template< typename T >
bool operator>(const Point<T>& p1, const Point<T>& p2) { 
   if (p1.x() != p2.x()) { return p1.x() > p2.x(); }
   else return p1.y() < p2.y();
}

template<typename T>
/* Calculate the squared euclidean distance between two points in 2D. */
T squared_distance(const Point<T>& p, const Point<T>& q) {
   return ((q.x()-p.x())*(q.x()-p.x()) + (q.y()-p.y())*(q.y()-p.y()));
}

template<typename T>
/* Calculate the euclidean distance between two points in 2D,
+  and return as double. Its probably a bad idea to call this if
+  T is not float, double or int.
*/
double squared_distance_dbl(const Point<T>& p1, const Point<T>& p2) {
   double a = static_cast<double>(p2.x() - p1.x());
   double b = static_cast<double>(p2.y() - p1.y());
   
   return (pow(a,2) + pow(b,2));
}


#endif
