/* Program der finder medianen i lineær tid, gennemsnitligt.
+  Se Cormen kap. 9.
+  Det der søges i er en vektor af punkter, dvs. det kan anskues som 
+  en n-dimensionel vektor. Der er n punkter, og hvert punkt har
+  m dimensioner. Denne udgave af randomized_select finder så
+  det i'te største element i en af de m dimensioner. 
+
+  Der er ogsp en funktion der finde den dimension med størst spredning,
+  dvs. hvor max-min er størst. I første omgang er den bundet til at
+  behandle 2-dimensionelle punkter.
+
+  Based on Cormen, Leiserson, Rivest, Stein: Introduction to Algorithms.
*/

#include <iostream>
#include <vector>
#include <cassert>
#include <cmath>


using namespace std;

template<typename T, typename Point>
class OrderStats {
   public:
      OrderStats() {}
      int getMedian(vector<Point>& vec, int p, int r, int dim);
      int randomized_select(vector<Point>& vec, int p, int r, int i, int dim);
      int calcDimension(vector<Point>& vec, int p, int r);

   private:
      int randomized_partition(vector<Point>& vec, int p, int r, int dim);
      int partition(vector<Point>& vec, int p, int r, int dim);
};


/* This is the median selector. p and r are start and end index in vec
+  for the range you are looking at. dim tells which dimension in Point 
+  you want to find the median for. */
template<typename T, typename Point>
int OrderStats<T,Point>::getMedian(vector<Point>& vec, int p, int r, int dim) {
   int half = static_cast<int>(ceil(double(r-p+1)/2.0));
   return randomized_select(vec, p, r, half, dim);
}


/* This selects the i'th order statistic. */
template<typename T, typename Point>
int OrderStats<T,Point>::randomized_select(vector<Point>& vec, int p, int r, int i, int dim) {

   if (p == r) { return p; }
   int q = randomized_partition(vec, p, r, dim);
   int k = q - p + 1;

   if (i == k)     { return q; }
   else if (i < k) { return randomized_select(vec, p, q-1, i, dim); }
   else            { return randomized_select(vec, q+1, r, i-k, dim); }   
}


/* Swaps vec[r] with a random other element, chosen from the range [p, r]. */
template<typename T, typename Point>
int OrderStats<T,Point>::randomized_partition(vector<Point>& vec, int p, int r, int dim) {

   double x = (r-p)*double(rand())/RAND_MAX;
   int i = int(x)+p;   // res = random number in [p, r].
  
   assert(i >= p && i <= r);

   Point tmp = vec[r];
   vec[r] = vec[i];
   vec[i] = tmp;
   
   return partition(vec, p, r, dim);
}

template<typename T, typename Point>
int OrderStats<T,Point>::partition(vector<Point>& vec, int p, int r, int dim) {

   T x = vec[r][dim];
   int i = p-1;
   for (int j=p; j<r; j++) {
      if (vec[j][dim] <= x) {
         ++i;
         Point tmp = vec[i];
         vec[i] = vec[j];
         vec[j] = tmp;
      }
   }

   Point tmp = vec[i+1];
   vec[i+1] = vec[r];
   vec[r] = tmp;
   
   return i+1;
}

template<typename T, typename Point>
int OrderStats<T,Point>::calcDimension(vector<Point>& vec, int p, int r) {

   T min[2]; 
   min[0] = vec[p][0]; 
   min[1] = vec[p][1];
   T max[2]; 
   max[0] = vec[p][0]; 
   max[1] = vec[p][1];
   
   for (int i=p+1; i<=r; i++) {
      if (min[0] > vec[i][0]) { min[0] = vec[i][0]; }
      if (max[0] < vec[i][0]) { max[0] = vec[i][0]; }
      if (min[1] > vec[i][1]) { min[1] = vec[i][1]; }
      if (max[1] < vec[i][1]) { max[1] = vec[i][1]; }
   }
   
   // Return the dimension with the biggest difference between max and min:
   if ((max[0]-min[0]) > (max[1]-min[1])) return 0;
   else                                   return 1;
}

