/* Class that holds the node (i.e. container) type
+  for use in a kdtree.
+
+  Template T is the type of the data that the kd-tree
+  is made of, e.g. int or double.
+  Template Point is the container for points. T is
+  the type of the individual coordinates inside Point.
+
+  Michael Neidhardt, Oct 2008.
+
+  This code is based on the description in J. L. Bentley,
+  "Multidimensional binary search trees used for associative
+  searching", Commun. ACM 18, 9 (1975), 509–517.
*/

#ifndef KDNODE_H
#define KDNODE_H

#include <iostream>
#include <vector>
#include <cassert>

using namespace std;

template<typename T, typename Point>
class KDNode {
public:
   KDNode();

   bool isLeaf()                    { return is_leaf; }
   KDNode<T,Point>* getLeftchild()  { return leftchild; }
   KDNode<T,Point>* getRightchild() { return rightchild; }
   Point* getLeafdata()             { return leafdata; }
   int getDimension()               { return dimension; }
   T getPartitionvalue()            { return partitionvalue; }

   void insertLeafdata(Point* data);
   void insertInternalNodedata(const int& d, const T& p);
   void insertLeftchild(KDNode<T,Point>* c)  { leftchild = c; }
   void insertRightchild(KDNode<T,Point>* c) { rightchild = c; }

private:
// is_leaf:        True if a leaf node, else false.
// dimension:      The dimension the data is split on in this internal node.
// partitionvalue: Value of this internal node.
// leftchild:      Left child of this internal node,
// rightchild:     Right child of this internal node.
// leafdata:       The data for this leaf node. Perhaps misleading to call
//                 this a Point, but its just a container for multidimensional
//                 data, such as a point in 2d.

   bool is_leaf;
   int dimension;              
   T partitionvalue;           
   KDNode<T,Point>* leftchild;
   KDNode<T,Point>* rightchild;
   Point* leafdata;       
};

// -------- Implementation here: ----------
/* Constructors: */
template<typename T, typename Point>
KDNode<T,Point>::KDNode() : is_leaf(false), dimension(-1), 
      partitionvalue(0), leftchild(0), rightchild(0) {}

/* End of constructors. */

template<typename T, typename Point>
void KDNode<T,Point>::insertLeafdata(Point* data) {
   leafdata = data;
   is_leaf = true;
}

template<typename T, typename Point>
void KDNode<T,Point>::insertInternalNodedata(const int& d, const T& p) {
   dimension = d;
   partitionvalue = p;
}

#endif
