/* Class used to build a kdtree and do range- and ANN-search.
+
+  The template params are:
+  T      = type of coordinate (coords of Node), and
+  Point  = type of container for points.
+  Traits = functions for selecting the median from a set and
+           function for finding which dimension 
+           contains the biggest spread.
+
+  Michael Neidhardt, Oct 2008.
+
+  This code is based on the description in J. L. Bentley,
+  "Multidimensional binary search trees used for associative
+  searching", Commun. ACM 18, 9 (1975), 509–517.
*/

#ifndef KDTREE_H
#define KDTREE_H

#include <iostream>
#include <vector>
#include <string>
#include <cassert>
#include "kdnode.h"


using namespace std;

template <typename T, typename Point, typename Traits>
class KDTree {
public:
   KDTree();
   void buildTree(vector<Point>& vec);
   Point* nn(Point* p);   // Finds the nearest neighbor of point p.
   void inOrderWalk() { inOrderWalk(root, 0); }

private:
   KDNode<T,Point>* buildTree(vector<Point>& vec, int p, int r, int counter);
   void rnn(KDNode<T,Point>* n);

   void inOrderWalk(KDNode<T,Point>* n, int depth);

   Traits mytraits;

   KDNode<T,Point>* root;

   Point* nntarget;
   Point* nnptnum;
   double nndist;
   int leavesvisited;
   
   T squared_distance(Point* p, Point* q) {
      return (((*q)[0]-(*p)[0])*((*q)[0]-(*p)[0]) + ((*q)[1]-(*p)[1])*((*q)[1]-(*p)[1]));
   }
};

// -------- Implementation here: ----------
/* Constructors: */
template <typename T, typename Point, typename Traits>
KDTree<T,Point,Traits>::KDTree() {}
/* End of constructors. */


template <typename T, typename Point, typename Traits>
void KDTree<T,Point,Traits>::buildTree(vector<Point>& vec) {
   root = buildTree(vec, 0, vec.size()-1, 0);
}


template <typename T, typename Point, typename Traits>
KDNode<T,Point>* KDTree<T,Point,Traits>::buildTree(vector<Point>& vec, int p, 
                                                   int r, int counter) {

   KDNode<T,Point>* node = new KDNode<T,Point>;

   
   if (p == r) { //if (r-p+1 <= cutoff) >>> If bucket size different from 1.
      node->insertLeafdata(&vec[p]);
   }
   else {
      int dim = mytraits.calcDimension(vec, p, r);
      int q =  mytraits.getMedian(vec, p, r, dim);

      node->insertInternalNodedata(dim, vec[q][dim]);
      node->insertLeftchild(buildTree(vec, p, q, counter+1));
      node->insertRightchild(buildTree(vec, q+1, r, counter+1));
   }

   return node;
}

template <typename T, typename Point, typename Traits>
Point* KDTree<T,Point,Traits>::nn(Point* p) {
   nntarget = p;
   nndist = DBL_MAX;
   leavesvisited=0;
   //std::cout << "Finding nn for [" << (*nntarget)[0] << "," << (*nntarget)[1] << "].\n";
   rnn(root);
   //std::cout << "Visited " << leavesvisited << " leaves.\n\n";
   return nnptnum;
}

template <typename T, typename Point, typename Traits>
void KDTree<T,Point,Traits>::rnn(KDNode<T,Point>* n) {

   if (n->isLeaf()) {
      ++leavesvisited;
      //std::cout << "In leaf: [" << (*n->getLeafdata())[0] << "," << (*n->getLeafdata())[1] << "]\n";
      
      // This needs to be here, or the function would always find the nn 
      // for a point p as p itself. I assume this was originally meant to look
      // for nearest neighbor of points not in the kd-tree.
      if (*n->getLeafdata() == *nntarget) { return; }
      
      double thisdist = squared_distance(n->getLeafdata(), nntarget);

      if (thisdist < nndist) {
         nndist = thisdist;
         nnptnum = n->getLeafdata();
      }
   }
   else {
      T val = n->getPartitionvalue();
      int dim = n->getDimension();
      T thisx = (*nntarget)[dim];

      T diff = thisx - val;
            
      if (diff < 0) {
         rnn(n->getLeftchild());
         if (nndist >= diff*diff) { rnn(n->getRightchild()); }
      }
      else {
         rnn(n->getRightchild());
         if (nndist >= diff*diff) { rnn(n->getLeftchild()); }
      }
   }
}


template <typename T, typename Point, typename Traits>
void KDTree<T,Point,Traits>::inOrderWalk(KDNode<T,Point>* n, int depth) {

   if ( ! n->isLeaf()) {
      inOrderWalk(n->getRightchild(), depth+4);
      
      for (int i=0; i<depth; i++) { std::cout << " "; }
      std::cout << n->getDimension() << "/" << n->getPartitionvalue() << "\n";

      inOrderWalk(n->getLeftchild(), depth+4);
   }
   else {
      for (int i=0; i<depth; i++) { std::cout << " "; }
      (n->getLeafdata())->print();
      std::cout << "\n";
   }
}

#endif
