/* This code is based on the description in K. Hinrichs, J. Nievergelt, and 
+  P. Schorn: "A sweep algorithm and its implementation: The all-nearest-neighbors 
+  problem revisited", Proceedings of the 14th international workshop on graph-
+  theoretic concepts in computer science, Springer-Verlag (1989), 442–457.
+
+  Written by Michael Neidhardt, June 2008.
*/
#ifndef ANNPLANESWEEP_H
#define ANNPLANESWEEP_H

#include <cassert>
#include <iostream>
#include <vector>
#include <queue>
#include <map>
#include <string>
#include "binutil.h"
#include "Point.h"
#include "Bisector.h"
#include "PSutil.h"
#include "QueueObject.h"
#include "priorityqueuewrapper.h"
#include "intervalfunctions.h"

using namespace std;

template< typename T >
class ANNPlaneSweep {
public:
   ANNPlaneSweep();
   ANNPlaneSweep(vector<Point<T> >& pset);   // C'tor with point set.
   vector<pair<Point<T>, Point<T> > > findAllNearestNeighbors();
   
private:
   typedef typename map<Bisector<T>, int, less<Bisector<T> > >::iterator mmap_iter;
   typedef multimap<Point<T>, Point<T>, less<Point<T> > > resultmmap;
   
   void findANN_Oneside(PQ<T>& xqueue);
   void findNN(Point<T>& p, Point<T>& nn, mmap_iter& lower, mmap_iter& upper);
   void updateYTable(PQ<T>& xqueue, 
                     Point<T>& p, 
                     Point<T>& nn, 
                     mmap_iter& iter, 
                     mmap_iter& result, 
                     bool goingUp, 
                     bool (intervalChecker<T>::*notInInterval)(Bisector<T>&, 
                                                               Bisector<T>&, 
                                                               Point<T>&)
                     );

   void checkForIntersection(PQ<T>& xqueue, Point<T>& p, mmap_iter& iter, bool both);
   void handleIntersection(PQ<T>& xqueue, Point<T>& p, vector<T>& is,
                           Bisector<T>& upperparent, Bisector<T>& lowerparent);

   vector<pair<Point<T>, Point<T> > > calcResult();
   intervalChecker<T> itvchk;

   vector<Point<T> > pointset;
   map<Bisector<T>, int, less<Bisector<T> > > ytable;
   resultmmap result;
;

   T minX, maxX, minY, maxY;
   int direction;
   int intersectioncount;  // Each intersection point encountered is added to 
                           // the prio-queue. If two such intersection points
                           // have identical coordinates, I want the one that
                           // is added first to also be taken out first.
   bool VERBOSE;   
};
template<typename T>
double squared_distance_dbl(const Point<T>& p1, const Point<T>& p2);

// -------- Implementation here: ----------

template< typename T >//---Constructor without initialization.
ANNPlaneSweep<T>::ANNPlaneSweep() {
   pointset.resize(0);
   VERBOSE=false;
}


template< typename T >//---Constructor with point set.
ANNPlaneSweep<T>::ANNPlaneSweep(vector<Point<T> >& pset) {
   pointset = pset;
   VERBOSE=false;
}


template<typename T>
vector<pair<Point<T>, Point<T> > > ANNPlaneSweep<T>::findAllNearestNeighbors() {
   PQ<T>* xqueue;

   direction = 1;
   xqueue = new minPQ<T>();
   findANN_Oneside(*xqueue);
   delete xqueue;
   ytable.clear();
   
   // std::cout << "Now the other direction. " << "\n";

   // NOW DOING OTHER DIRECTION --- \n";
   // Doing the 'other direction' simply means multiplying all x-coords by -1.
   direction = -1;
   xqueue = new maxPQ<T>();
   findANN_Oneside(*xqueue);
   delete xqueue;
   ytable.clear();
      
   return calcResult();
}


template< typename T >//---Finds those neighbors that are to the left/right.
void ANNPlaneSweep<T>::findANN_Oneside(PQ<T>& xqueue) {

   minX = direction*pointset[0].x(); maxX = direction*pointset[0].x();
   minY = pointset[0].y(); maxY = pointset[0].y();
   intersectioncount = -1;

   // First insert x-values in priority queue. The info-part
   // is the index of the point in the vector pointset. I also find
   // min and max for the coordinates.

   for (unsigned int i=0; i<pointset.size(); i++) {
      Point<T> tmp(direction*pointset[i].x(), pointset[i].y());
      QueueObject<T> qo(tmp, 0);
      xqueue.push(qo);
      minX = std::min(minX, direction*pointset[i].x());
      minY = std::min(minY, pointset[i].y());
      maxX = std::max(maxX, direction*pointset[i].x());
      maxY = std::max(maxY, pointset[i].y());
   }


   /* ----------Processing the first two points - done manually since--- */
   /*           it makes life a lot easier:                              */
   QueueObject<T> first = xqueue.top();
   xqueue.pop();
   QueueObject<T> second = xqueue.top();
   xqueue.pop();
   
   Bisector<T> firstBS(first.point, second.point, minY, maxY);
   ytable[firstBS] = 0;


   first.point.init(first.point.x()*direction, first.point.y());
   second.point.init(second.point.x()*direction, second.point.y());

   result.insert(pair<Point<T>,Point<T> >(first.point,second.point));
   result.insert(pair<Point<T>,Point<T> >(second.point,first.point));
   /* -----------Done processing the first two points------------------ */

  
   while ( ! xqueue.empty()) {
      QueueObject<T> qo = xqueue.top();
      xqueue.pop();

      if (qo.index >= 0) {          // This is a data point.
         mmap_iter lower, upper;
         Point<T> nn;

         findNN(qo.point, nn, lower, upper);
         Point<T> tmp1(qo.point.x()*direction, qo.point.y());
         Point<T> tmp2(nn.x()*direction, nn.y());
         result.insert(pair<Point<T>,Point<T> >(tmp1, tmp2));


         // The bisector(s) that is (are) inserted are pointed to by these two iterators:
         mmap_iter upresult, downresult;
         
         updateYTable(xqueue, qo.point, nn, upper, upresult,
                      true, &intervalChecker<T>::notInIntervalUP);
         updateYTable(xqueue, qo.point, nn, lower, downresult,
                      false, &intervalChecker<T>::notInIntervalDOWN);
         
         /* Below are 3 cases (there is always one or two new bisectors
         +  inserted as a result of the above 2 calls to updateYTable):
         +  1: Update upwards inserted no bisector, so check the new bisector 
         +     inserted during update downwards for intersection with its upper and
         +     lower neighbors.
         +  2: Update downwards inserted no bisector, so check the new bisector
         +     inserted during update upwards for intersection with its upper 
         +     and lower neighbors.
         +  3: Update up and down both inserted a bisector, so check the bisector
         +     inserted during update up for intersection with its upper neighbor,
         +     and check the bisector inserted during update downwards for intersection
         +     with its upper and its lower neighbors. Note that in this case, the upper
         +     neighbor of the new bisector is the one inserted during update upwards.
         */
         if (upresult == ytable.end()) {
            checkForIntersection(xqueue, qo.point, downresult, true);
         }
         else if (downresult == ytable.end()) {
            checkForIntersection(xqueue, qo.point, upresult, true);
         }
         else {
            checkForIntersection(xqueue, qo.point, upresult, false);
            checkForIntersection(xqueue, qo.point, downresult, true);
         }
      }
      else {                        // This is an intersection point.
         mmap_iter upper = ytable.find(qo.upperparent);
         mmap_iter lower = ytable.find(qo.lowerparent);


         // If both "parent -bisectors" of this intersection point 
         // still exist in ytable, I erase them abd insert their replacement.
         // Under certain circumstances, ytable.find will find the wrong
         // bisector, and therefore I also check if the bisectors actually found
         // (i.e. upper and lower) are in fact equal to upperparent and lowerparent:
         if (upper != ytable.end() && lower != ytable.end()) {
            Bisector<T> replacement(qo.upperparent.getUpperpoint(),
                                    qo.lowerparent.getLowerpoint(),
                                    minY, maxY);

            ytable.erase(qo.upperparent);
            ytable.erase(qo.lowerparent);
            pair<mmap_iter,bool> retval = 
               ytable.insert(pair<Bisector<T>,int>(replacement,0));
            
            checkForIntersection(xqueue, qo.point, retval.first, true); 
         }
      }
   }
}

template<typename T >// --- Finds the nearest neighbor to the left/right
void ANNPlaneSweep<T>::findNN(Point<T>& p, Point<T>& nn, 
                              mmap_iter& lower, mmap_iter& upper) {

   // Horizontal bisector through p, the point we are looking for.
   Bisector<T> hb(p, minY, maxY);

   upper = ytable.upper_bound(hb);
   lower = upper;

   if (upper != ytable.end()) {
      nn = (upper->first).getLowerpoint();
   }
   
   if (lower == ytable.end()) {
      if ( ! ytable.empty()) {
         lower--;
         nn = (lower->first).getUpperpoint();
      }
   }
   else if (lower == ytable.begin()) {
      lower = ytable.end();
   }
   else {
      lower--;
      nn = (lower->first).getUpperpoint();
   }
}


template<typename T>
void ANNPlaneSweep<T>::updateYTable(PQ<T>& xqueue,
                                    Point<T>& p, 
                                    Point<T>& nn, 
                                    mmap_iter& iter,
                                    mmap_iter& result,
                                    bool goingUp,
                                    bool (intervalChecker<T>::*notInInterval)
                                       (Bisector<T>&, Bisector<T>&, Point<T>&)) {


   // If result == ytable.end(), nothing was inserted. If, on the other hand,
   // a bisector was inserted, result will contain an iterator pointing to it.
   result=ytable.end();

   vector<Bisector<T> > trashcan(0);

   Bisector<T> currentBS(p, nn, minY, maxY);
   Bisector<T> nextBS;

   if (iter != ytable.end()) { 
      nextBS = iter->first;
   }

   // The condition here is: iter must not be at end() and one of the following
   // must also hold:
   //    a) currentBS is vertical
   //    b) currentBS doesn't intersect sweep line in the interval
   //       between p.y() and nextBS' y-value.
   while (iter != ytable.end() &&
          (currentBS.get_b() == 0 ||
          (itvchk.*notInInterval)(currentBS, nextBS, p))
          ) {

      Point<T> q = (goingUp) ? nextBS.getUpperpoint() : nextBS.getLowerpoint();

      if (goingUp) {
         iter++;
      }
      else {
         // If iter at begin(), we can't decrement, so signal that it is at end().
         if (iter == ytable.begin()) { iter = ytable.end(); }
         else                        { iter--; }
      }

      // The iterator iter can be messed up if I erase here,
      // so I save it in trashcan till after the loop we're in.
      trashcan.push_back(nextBS);  

      if (iter != ytable.end()) { nextBS = iter->first; }
      else                      { nextBS = Bisector<T>(); }

      currentBS = Bisector<T>(p, q, minY, maxY);
   }

   for (unsigned int i=0; i<trashcan.size(); i++) {
      ytable.erase(trashcan[i]);
   }


   // Possibly insert the new bisector:
   if (iter != ytable.end() || 
       (currentBS.get_b() == 0 && goingUp) ||
       (currentBS.get_b()!=0 && currentBS.getYValueAsDouble(p.x()) >= p.y() && goingUp) ||
       (currentBS.get_b()!=0 && currentBS.getYValueAsDouble(p.x()) <= p.y() && ! goingUp)) {

      pair<mmap_iter,bool> retval = 
         ytable.insert(pair<Bisector<T>,int>(currentBS,0));
      
      /*
      if (retval.second == false) {
         std::cerr << "Panic - Bisector exists in ytable. Shouldnt happen.\n";
         currentBS.print("Dublet: ");
         result = retval.first;
      }
      else { result = retval.first; }
      */
      result = retval.first;
   }
}


/* Checks if new bisector(s) intersect its one or two neighbors in ytable.
   If arg. 'both' is false, the bisector pointed to by iter is checked only for 
   intersection with its upper neighbor.
   If arg. 'both' is true, the bisector pointed to by iter is checked
   for intersection with its upper as well as its lower neighbor.
   This is because 2 new bisectors can intersect each other, as well as the neighbor
   on the 'outside'. I.e. the bisector inserted during update upwards, can
   intersect its upper neighbor, and the bisector inserted during update downwards
   can intersect its lower neighbor, and the 2 new bisectors can intersect each other.
*/
template<typename T>
void ANNPlaneSweep<T>::checkForIntersection(PQ<T>& xqueue, 
                                             Point<T>& p, 
                                             mmap_iter& iter,
                                             bool both) {

   Bisector<T> currentBS(iter->first);
   iter++;

   if (iter != ytable.end()) {
      Bisector<T> successor(iter->first);
      vector<T> is = calcIntersection_v2(successor, currentBS);

      if (is[2] != 0) handleIntersection(xqueue, p, is, successor, currentBS);
   }

   if (both) {
      iter--;        // Now iter points to currentBS.
         
      if (iter != ytable.begin()) { // I.e. if currentBS is not the first:
         iter--;         // This now points to the predecessor of currentBS.
         Bisector<T> predecessor(iter->first);
         vector<T> is = calcIntersection_v2(currentBS, predecessor);
         if (is[2] != 0) handleIntersection(xqueue, p, is, currentBS, predecessor);
      }
   }

}


template<typename T>
void ANNPlaneSweep<T>::handleIntersection(PQ<T>& xqueue, 
                                          Point<T>& p,
                                          vector<T>& is, 
                                          Bisector<T>& upperparent, 
                                          Bisector<T>& lowerparent) {

   double is_x = static_cast<double>(is[0])/static_cast<double>(is[2]); 
   double is_y = static_cast<double>(is[1])/static_cast<double>(is[2]);

// TEST for intersection point is = -7353 7155 54.
/*
if (is[0]==-7353 && is[1]==7155 && is[2]==54) {
   p.print("p= "); cout << "\n";
   
   binutil bu;
   cout << "BinRep of " << is[0] << ": ";
   bu.showBinaryRep(is[0]);
   bu.doubleprint(is[0]);
   cout << "\nBinRep of " << (p.x()*is[2]) << ": ";
   bu.showBinaryRep(p.x()*is[2]);
   cout << "\n";
   cout << "BinRep of " << (is[0]/is[2]) << ": ";
   bu.showBinaryRep(is[0]/is[2]);
   cout << "BinRep of " << p.x() << ": ";
   bu.showBinaryRep(p.x());
   cout << "\n";
}
*/

   //if (is[0] >= p.x()*is[2] && is[0] <= maxX*is[2]) { 
     // if (is[1] >= minY*is[2] && is[1] <= maxY*is[2]) {
   if (is_x >= p.x() && is_x <= maxX) { 
      if (is_y >= minY && is_y <= maxY) {
         // Add to xqueue.
         QueueObject<T> qo(Point<T>(is_x, is_y), -1, 
                           upperparent, lowerparent);
         xqueue.push(qo);
         
      }
   }
}


template<typename T>// --- Merges and prints out result:
vector<pair<Point<T>, Point<T> > > ANNPlaneSweep<T>::calcResult() {
   typedef typename multimap<Point<T>, Point<T>, less<Point<T> > >::iterator ppmap_iter;
   vector<pair<Point<T>, Point<T> > > finalresult(0);


   ppmap_iter mainiter(result.begin()); 

   /* Loop through all points with their accompanying (suggested) neighbors. */
   while (mainiter != result.end()) {
      // Calculate distance between the first point and its NN:
      double distance = squared_distance_dbl(mainiter->first, mainiter->second);
      // Pick out the first point and its NN:
      Point<T> currentpoint(mainiter->first);
      Point<T> nearestneighbor(mainiter->second);  // Suggested NN only!

      
      // Find all points equal to currentpoint (there is at least 1):
      pair<ppmap_iter, ppmap_iter> rangeiter = result.equal_range(currentpoint);

      // Each point has two suggestions - one NN to the left, and one to the right.
      // Their might also be duplicate points. This gets them all, and lastly finds
      // the one that is nearest:
      for (ppmap_iter subiter=rangeiter.first; subiter!=rangeiter.second; ++subiter) {
         const Point<T> p1 = subiter->first;
         const Point<T> p2 = subiter->second;
         //T tmp = squared_distance(p1, p2);
         double tmp = squared_distance_dbl(subiter->first, subiter->second);

         
         // I dont want more than one, so i use < (less than) and 
         // not <= (less than or equal).
         if (tmp < distance) {
            nearestneighbor = subiter->second;
            distance = tmp;
         }
      }

      finalresult.push_back(pair<Point<T>, Point<T> >(currentpoint, nearestneighbor));
      
      result.erase(mainiter->first);
      mainiter = result.begin();
   }

   return finalresult;
}

#endif
