/* Class that is used to hold an object of type Bisector.
+  This type of object is created by two points.
+  From these 2 points are created the equation for the line that bisects them.
+  Thisn equation has 3 coefficients, a, b, c, characterising a line in 2D.
+  ax + by = c.
+
+  Originally made for the ANN Plane sweep algorithm.
+  Michael Neidhardt, May 2008.
*/

#ifndef BISECTOR_RAT_H
#define BISECTOR_RAT_H

#include "Point.h"
#include <cassert>
#include <iostream>
#include <string>
#include <LEDA/numbers/rational.h>

using namespace std;

template <typename T>
class Bisector {
public:
   Bisector();
   Bisector(Point<T> p, T pminY_, T pmaxY_);
   Bisector(Point<T> p1, Point<T> p2, T pminY_, T pmaxY_);
   Bisector(const Bisector<T>& bisector);

   T getYValue(T x) const;

   Point<T> getUpperpoint() const { return upperpoint; }
   Point<T> getLowerpoint() const { return lowerpoint; }

   // This returns whichever of the two x-coordinates is greater.
   T getmaxX() const { return std::max(upperpoint.x(), lowerpoint.x()); }
   // This returns whichever of the two x-coordinates is less.
   T getminX() const { return std::min(upperpoint.x(), lowerpoint.x()); }

   // See below:
   T getPointset_minY() const { return pminY; }
   T getPointset_maxY() const { return pmaxY; }

   T get_a() const { return a; }
   T get_b() const { return b; }
   T get_c() const { return c; }

   bool isDefined() const { return defined; }

   void print() const { print(""); }
   void print(string info) const;
   void print(string info, T x) const;

   Bisector<T>& operator=(const Bisector<T>& rightarg);

private:
   void setCoefficients();

   Point<T> upperpoint;
   Point<T> lowerpoint;
   T a, b, c;
   bool defined;
   
   T pminY, pmaxY;
   // These two hold the minimum and maximum y-value used in the pointset currently
   // being processed. Messy, but a simple way to solve the problem
   // with the order of bisectors. I use them in operator <, for determining
   // which of two bisectors is the 'smaller'. If the both intersect the sweep line
   // below minY or above maxY, the bisector with the smallest maxX is the smaller.
   // This simply means that the bisector furthest to the left is less than the one
   // to its right, i.e. the one closest to the sweep line is greater than the other.
   // This is symmetric for bisectors which intersect sweep line above maxY: The one
   // whose maxX is closest to the sweep line is smaller than the other.

};

// ---------- Non member stuff -----------------------------
template< typename T >
bool operator==(const Bisector<T>& bis1, const Bisector<T>& bis2);
template< typename T >
bool operator!=(const Bisector<T>& bis1, const Bisector<T>& bis2);
template< typename T >
bool operator<(const Bisector<T>& bis1, const Bisector<T>& bis2);
template< typename T >
std::vector<T> calcIntersection_v1(Bisector<T>& upper, Bisector<T>& lower);
template< typename T >
std::vector<T> calcIntersection_v2(Bisector<T>& upper, Bisector<T>& lower);


// -------- Implementation here: ----------
template< typename T >
Bisector<T>::Bisector() : a(0),b(0),c(0),defined(false),pminY(0),pmaxY(0) {}

template< typename T >
// C'tor with one point. This creates a horizontal bisector
// through the point. Strictly speaking this is not part of the class,
// but I have put it here to make life easier when doing ANNPlanesweep.
Bisector<T>::Bisector(Point<T> p, T pminY_, T pmaxY_) {
   upperpoint.init(p.x(), p.y()+1);
   lowerpoint.init(p.x(), p.y()-1);

   a = 0;
   b = 1;
   c = p.y();
   defined=true;
   pminY = pminY_;
   pmaxY = pmaxY_;
}

template< typename T >
/* C'tor with two points (the generators for this bisector):
+  If the two points have same y-coord, the point with greatest 
+  x-coord becomes lower point. This is because I only do a sweep
+  from left to right, and a vertical bisector is always 'higher'
+  than a non-vertical (this is primarily or perhaps only used in
+  used in PlaneSweep ANN).
*/
Bisector<T>::Bisector(Point<T> p1, Point<T> p2, T pminY_, T pmaxY_) {
   if (p1.y() > p2.y()) {
      upperpoint = p1;
      lowerpoint = p2;
   }
   else if (p1.y() < p2.y()) {
      upperpoint = p2;
      lowerpoint = p1;
   }
   else if (p1.x() > p2.x()) {    // From here on down, p1.y == p2.y.
      upperpoint = p2;
      lowerpoint = p1;
   }
   else {
      upperpoint = p1;
      lowerpoint = p2;
   }

   setCoefficients();
   defined=true;
   pminY = pminY_;
   pmaxY = pmaxY_;
}

template< typename T >
// Copy c'tor.
Bisector<T>::Bisector(const Bisector<T>& bisector) {

   upperpoint = bisector.getUpperpoint();
   lowerpoint = bisector.getLowerpoint();
   a = bisector.get_a();
   b = bisector.get_b();
   c = bisector.get_c();
   
   defined = bisector.isDefined();
   pminY = bisector.getPointset_minY();
   pmaxY = bisector.getPointset_maxY();
}

template< typename T >
T Bisector<T>::getYValue(T x) const {
   return (-a*x + c)/b;
}


template< typename T >
void Bisector<T>::print(string info) const {
   cout << info ;
   if (!defined) {
      std::cout << "Bisector not defined.\n";
   }
   else {
      upperpoint.print();
      lowerpoint.print("-");
      cout << "  [a b c] = [" << a << " " << b << " " 
           << c << " " << "]\n";
   }
}

template< typename T >
void Bisector<T>::print(string info, T x) const {
   cout << info ;
   if (!defined) {
      std::cout << "Bisector not defined.\n";
   }
   else {
      upperpoint.print();
      lowerpoint.print("-");
      cout << "  [a b c] = [" << a << " " << b << " " 
           << c << " " << "] f(" << x << ")=" 
           << getYValue(x) << "\n";
   }
}

template< typename T >
Bisector<T>& Bisector<T>::operator=(const Bisector<T>& rightarg) {
   upperpoint = rightarg.getUpperpoint();
   lowerpoint = rightarg.getLowerpoint();
   a = rightarg.get_a();
   b = rightarg.get_b();
   c = rightarg.get_c();
      
   defined = rightarg.isDefined();

   pminY = rightarg.getPointset_minY();
   pmaxY = rightarg.getPointset_maxY();

   return *this;
}


template< typename T >
void Bisector<T>::setCoefficients() {
   a = 2*(upperpoint.x()-lowerpoint.x());
   b = 2*(upperpoint.y()-lowerpoint.y());
   c = upperpoint.x()*upperpoint.x()-lowerpoint.x()*lowerpoint.x() + 
       upperpoint.y()*upperpoint.y()-lowerpoint.y()*lowerpoint.y();
}


/* ----- Non member functions: ------------ */
template< typename T >
bool operator==(const Bisector<T>& bis1, const Bisector<T>& bis2) { 
   return (bis1.getUpperpoint() ==  bis2.getUpperpoint() && 
           bis1.getLowerpoint() ==  bis2.getLowerpoint());
}

template< typename T >
bool operator!=(const Bisector<T>& bis1, const Bisector<T>& bis2) { 
   return (bis1.getUpperpoint() !=  bis2.getUpperpoint() ||
           bis1.getLowerpoint() !=  bis2.getLowerpoint());
}

/* NB: There is no '>' (greater than), since it is not used.
+
+  When comparing two lines, obviously it needs to be done at a specific
+  x-coordinate. Since with STL I cannot give a third parameter to the compare
+  function used in the map/multimap, i.e. an x-value, I resort to taking 
+  the extreme x-value in the two bisectors involved in the comparison. 
+  This works, since the bisector with the correct x-value is always
+  part of the comparison. I.e. when searching for a bisector, the one
+  we look for is propagated on in the recursion until we find the answer.
*/

template< typename T >
bool operator<(const Bisector<T>& bis1, const Bisector<T>& bis2) {
   // An undefined bisector is always greater than anything else.
   if (! bis2.isDefined())      return true;
   else if (! bis1.isDefined()) return false;

   T maxX1 = bis1.getmaxX();        // The greater of the 2 xcoords of bis1.
   T maxX2 = bis2.getmaxX();        // The greater of the 2 xcoords of bis2.
   const T xvalue = std::max(maxX1, maxX2);

   T a1 = bis1.get_a(); T b1 = bis1.get_b(); T c1 = bis1.get_c();
   T a2 = bis2.get_a(); T b2 = bis2.get_b(); T c2 = bis2.get_c();

   T p = xvalue.numerator();
   T q = xvalue.denominator();

   // First take care of situations with vertical bisector(s):
   if (b1 == 0 && b2 > 0) { 
      return false;
   }
   else if (b1 > 0 && b2 == 0) { 
      return true;
   }
   else if (b1 == 0 && b2 == 0) {
      T v1 = (abs(bis1.getLowerpoint().x() - bis1.getUpperpoint().x()) + 
              2*leda::min(bis1.getLowerpoint().x(), bis1.getUpperpoint().x()));

      T v2 = (abs(bis2.getLowerpoint().x() - bis2.getUpperpoint().x()) + 
              2*leda::min(bis2.getLowerpoint().x(), bis2.getUpperpoint().x()));

      return (v1 > v2);
      // If vertical bisector, the bisector that lies furthest to the right is 'less than'
      // the other. Furthest to the right means that the x-coord of the bisector (!) is
      // greater than that of the other.
   }
   else { // Neither bisector is vertical if I get here, so calculate the actual y-values 
      // for the bisectors on the sweep line. Needed to find out if we're in the 'rectangle'
      // or not. The 'rectangle' is the one defined by the sweep line as leftmost x, max.x-coord
      // of the pointset as rightmost x, and the minimum and maximum y-coords of the pointset.
      T yval1 = (c1-a1*xvalue)/b1;
      T yval2 = (c2-a2*xvalue)/b2;

      // The 'y-values' used when comparing - see the article:
      T yval1_cmp = b2*(q*c1-p*a1);
      T yval2_cmp = b1*(q*c2-p*a2);

      // As described above, these are minimum x- and y-values for
      // the pointset that is being processed.
      T pminY = bis1.getPointset_minY();
      T pmaxY = bis1.getPointset_maxY();
      
      if (yval1_cmp != yval2_cmp) {
         if (yval1 < pminY && yval2 < pminY)
            return bis1.getmaxX() < bis2.getmaxX();
         else if (yval1 > pmaxY && yval2 > pmaxY)
            return bis1.getmaxX() > bis2.getmaxX();
         else
            return yval1_cmp < yval2_cmp;
      }
      else {
         if (a1/b1 != a2/b2)      
            return (a1/b1 < a2/b2);
         else if (maxX1 != maxX2)
            return maxX1 < maxX2;
         else
            return bis1.getminX() < bis2.getminX();
      }
   }
}


/* Calculates the intersection of two neighbouring bisectors.
+ This is based on the formula in my 'formelsamling' ;-)
+ Returns 3 numbers: v[0] = x, v[1] = y, v[2] = d. The intersection
+ point is (x/d, y/d). 
*/
template< typename T >
std::vector<T> calcIntersection_v1(Bisector<T>& upper, Bisector<T>& lower) {
   vector<T> v(3);
   
   T a1 = upper.get_a();
   T b1 = upper.get_b();
   T c1 = upper.get_c();

   T a2 = lower.get_a();
   T b2 = lower.get_b();
   T c2 = lower.get_c();


   v[0] = c1*b2 - c2*b1;
   v[1] = a1*c2 - a2*c1;
   v[2] = a1*b2 - a2*b1;

   return v;
}

/* Calculates the intersection of two neighbouring bisectors.
+  This is based on the formula in the article. It requires the user
+  to supply the two bisectors as upper first and lower second.
+  The two bisectors need to have a point in common, and I need to 
+  know which it is. This is probably faster than my version above,
+  and at least it requires less precision.
+  I.e. the first must be at a higher y-value (at the current x) than the second.
+  Returns 3 numbers: v[0] = X, v[1] = Y, v[2] = D. 
+  The intersection point is (X/D, Y/D). 
*/
template< typename T >
std::vector<T> calcIntersection_v2(Bisector<T>& upper, Bisector<T>& lower) {
   vector<T> v(3);
   
   Point<T> p1 = upper.getUpperpoint();
   Point<T> p2 = upper.getLowerpoint();

   Point<T> p3 = lower.getLowerpoint();
   //Point<T> p4 = lower.getUpperpoint();  // Supposed to be equal to p2.

   v[0] = (p1.x()*p1.x() + p1.y()*p1.y())*(p2.y() - p3.y()) + 
          (p2.x()*p2.x() + p2.y()*p2.y())*(p3.y() - p1.y()) + 
          (p3.x()*p3.x() + p3.y()*p3.y())*(p1.y() - p2.y());

   v[1] = (p1.x()*p1.x() + p1.y()*p1.y())*(p3.x() - p2.x()) + 
          (p2.x()*p2.x() + p2.y()*p2.y())*(p1.x() - p3.x()) + 
          (p3.x()*p3.x() + p3.y()*p3.y())*(p2.x() - p1.x());

   v[2] = 2*(p1.x()*p2.y() - p2.x()*p1.y() + p2.x()*p3.y() - 
             p3.x()*p2.y() + p3.x()*p1.y() - p1.x()*p3.y());
   return v;
}

#endif
