/* Program that triangulates a point set and calculates
+  ANN on the basis of the triangulation.
+  
+  Exact version.
+
+  Michael Neidhardt, June 2008.
*/

#include <fstream>
#include <string>
#include <CGAL/basic.h>
#include <CGAL/Gmpq.h>
#include <CGAL/Cartesian.h>
#include <CGAL/Delaunay_triangulation_2.h>
#include <CGAL/Triangulation_hierarchy_2.h>
#include "annutil.h"
#include "Datasource2.h"
#include "ANNComparator2.h"

typedef CGAL::Gmpq                                             Coord_type;
typedef CGAL::Cartesian<Coord_type>                            K;
typedef K::Point_2                                             Point;
typedef CGAL::Triangulation_vertex_base_2<K>                   Vbb;
typedef CGAL::Triangulation_hierarchy_vertex_base_2<Vbb>       Vb;
typedef CGAL::Triangulation_face_base_2<K>                     Fb;
typedef CGAL::Triangulation_data_structure_2<Vb,Fb>            Tds;
typedef CGAL::Delaunay_triangulation_2<K,Tds>                  Dtr;
typedef CGAL::Triangulation_hierarchy_2<Dtr>                   Delaunay;
/* ---------------------------------------------------------------- */

typedef Delaunay::Vertex_iterator           Vertex_iterator;
typedef Delaunay::Vertex_circulator         Vertex_circulator;
typedef Delaunay::Vertex_handle             Vertex_handle;


double fillItUp(Delaunay& dt, int argc, char** argv);
void findAllNearestNeighbors(Delaunay& dt, vector<Point>& pointset,
                             vector<pair<Point, Point> >& ann);



int main(int argc, char** argv) {
   Delaunay dt;
   std::string progname = "CGALGMP";

   if (argc == 1) {
      cerr << progname 
           << "\nSyntax: " << argv[0] << " pointtype [filename|numOfPoints] "
           <<                            " [precision] [verbose] [check]\n\n"
           << "   pointtype=0 means the following arg is filename with points.\n" 
           << "   precision: must be in [1,31]. Decides precision of random numbers.\n"
           << "   verbose: 0=quiet, 1=one-line summary, 2=all points listed.\n"
           << "   check:   1=check against brute force results. 0=no check.\n\n";

      exit(0);
   }

   double dttime = fillItUp(dt, argc, argv);
   
   double anntime = annutil::microsec();
   vector<Point> pointset(0);    // I need this for checking the result.
   vector<pair<Point, Point> > ann(0);
   findAllNearestNeighbors(dt, pointset, ann);
   anntime = annutil::microsec(anntime);

   if (argc > 5 && atoi(argv[5]) == 1) {
      double chktime = annutil::microsec();
      ANNComparator<Coord_type,Point> comp;
      int errors = comp.doComparison(ann, pointset, atoi(argv[4]));
      chktime = annutil::microsec(chktime);

      if (atoi(argv[4]) > 0) {
         annutil::printResultWithCheck(progname, pointset.size(), ann.size(), atoi(argv[3]), errors,
                                 dttime, anntime, chktime);
      }
      exit(errors);
   }
   else {
      if (argc > 4 && atoi(argv[4]) > 0) {
         annutil::printResultWithoutCheck(progname, pointset.size(), ann.size(), atoi(argv[3]),
                                          dttime, anntime);
      }
      exit(0);
   }

   return 0;
}

// This just fills up a vector with points, and loads that into the dt.
double fillItUp(Delaunay& dt, int argc, char** argv) {

   Datasource2<Coord_type, Point> dsrc;
   vector<Point> pointset(0);

   if (atoi(argv[1]) == 0) {
      pointset = dsrc.readPointset(argv[2]);
   }
   else {
      std::cerr << "Input must come from a file. First arg. should be 0.\n";
      exit(1);
   }

   // Dont want to include point creation in the time taken for Delaunay triangulation:
   double dttime = annutil::microsec();

   std::vector<Point>::iterator iter;
   
   for (iter = pointset.begin(); iter != pointset.end(); iter++) {
      dt.insert(*iter);
   }

   return annutil::microsec(dttime);
}


/* This finds the nearest neighbor for each point in triangulation.
+  The pointset created in fillItUp might contain duplicates. They are discarded
+  when the triangulation is constructed, so I create the vector pointset here,
+  with the actual points used in dt. Otherwise I will get errors when comparing
+  my result here with that of the brute force method.
+  ann contains pairs of Points: For each point in the triangulation, I record its NN.
+  pointset contains just points, i.e. its actually a copy of the first element of 
+  each pair in ann... Anyway, pointset is used when doing a brute force, for checking.
*/
void findAllNearestNeighbors(Delaunay& dt, vector<Point>& pointset, 
                             vector<pair<Point, Point> >& ann) {
   Vertex_iterator vi;

   // Outer loop runs over all nodes of the triangulation.
   for (vi = dt.finite_vertices_begin(); vi != dt.finite_vertices_end(); vi++) {
      Vertex_circulator vc = dt.incident_vertices(vi);
      if (dt.is_infinite(vc)) { ++vc; }
      Point firstIncidentPoint(vc->point());
      Point nn(firstIncidentPoint);
      Coord_type distance = squared_distance(vi->point(), vc->point());

      // Inner loop runs over all edges for each node.
      do {
         if ( ! dt.is_infinite(vc)) {   //OR THIS?: if (vc != dt.infinite_vertex())
            Coord_type distance2 = squared_distance(vi->point(), vc->point());
            if (distance2 < distance) {
               nn = vc->point();
               distance = distance2;
            }
         }
         ++vc;
      } while (vc->point() != firstIncidentPoint);

      pointset.push_back(vi->point());
      ann.push_back(pair<Point, Point>(vi->point(), nn));
   }
}
